/* nc100em, an Amstrad NC100/NC200 emulator.
 * Copyright (C) 1995-2022 Russell Marks and Ian Collier.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 *
 * common.c - bits common to all versions.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include <SDL2/SDL_timer.h>

#include "common.h"
#include "z80.h"
#include "pdrom.h"

#ifdef NC200_DISK_SUPPORT
#include <765.h>
#endif


/* name of file for printer output (appended to if any output received) */
#define PRINTER_OUTPUT	"printout.dat"

/* disk image filename for NC200 support */
#define NC200_DISK_IMAGE "nc200.dsk"

#define libdir(x) (x)


/* the memory is (up to) 512k ROM and 128k RAM */
unsigned char mem[(512+128)*1024];
/* ROM-like `memory' used when no mem card is present */
unsigned char nocardmem[16*1024];
unsigned char *cardmem=NULL;	/* malloc'd or mmap'd PCMCIA card mem */
unsigned char *dfile;
/* boots with all ROM0 */
unsigned char *memptr[4]={mem,mem,mem,mem};
unsigned char lastpageout[4]={0,0,0,0};	/* last outs to paging ports */

/* In 2022 I tested a real NC200 as running at 4.629MHz, but I
 * think that might be reflecting clock drift on old hardware,
 * as it's only about 0.5% different. So I've stuck with the old
 * 4.606MHz I got from testing an NC100 in the 90s or thereabouts.
 */
unsigned long tstates=0,tsmax=46060;

int memattr[4]={0,0,0,0};	/* all ROM at boot */

volatile int signal_int_flag=0;
volatile int exit_program_flag=0;
int scrn_freq=1;
int default_scale=SCALE;

int card_size=0;		/* size of pcmcia card, or 0 if none */
int card_status=0;		/* as read from port 0xA0 */
int irq_status=0xff;		/* as read from port 0x90 */
int irq_mask=0;			/* as set via port 0x60 */
int scrnaddr=0xf000;		/* slightly nicer than assuming zero :-) */
int do_nmi=0;
int force_full_redraw=1;
int force_pd_rom=0;
int using_pd_rom=0;
int allow_serial=0;
int alt_colour_msg=0;	/* alter ROM colour msgs (e.g. "RED" -> "LEFT") */
int wait_for_sec=0;	/* wait for a second before starting up; gives you
                           a chance to do things like function/stop/del */
int invert_video=0;

unsigned char keyports[10]={0,0,0,0,0, 0,0,0,0,0};

struct termios old_tty_st;
int tty_fd=0;				/* use stdin */
unsigned char seroutbuf[16384];		/* won't need this many - defensive */
int seroutpos=0;

static char *binfile_name=NULL;
static int do_munmap=1;
int write_ram_on_close=0;

FILE *printout=NULL;

/* defaults to NC100 */
int nc200=0;
int nc150=0;


#ifdef NC200_DISK_SUPPORT
FDC_PTR nc200_fdc;
FDRV_PTR nc200_drive,nc200_drive_dummy;
int diskint=0;
#endif


/* the card_status flag never changes after this */
void set_card_status(void)
{
card_status=0x31; /* no batteries low, and parallel always ACK and !BUSY */
		  /* (note that nciospec.doc documents BUSY wrongly!) */
if(nc200)
  {
  card_status=0x10;	/* the '200 has a different take on it */
#ifndef NC200_DISK_SUPPORT
  /* if no disk support, pretend the batteries aren't up to it :-) */
  card_status|=1;
#endif
  }

if(card_size==0) card_status|=0x80;	/* set bit 7 if no card present */
}


void loadrom(unsigned char *x)
{
FILE *in;
char *rom_file="nc100.rom";

if(nc150) rom_file="nc150.rom";
if(nc200) rom_file="nc200.rom";

if(!force_pd_rom &&
   (in=fopen(libdir(rom_file),"rb"))!=NULL)
  {
  fread(x,1024,(nc150 || nc200)?512:256,in);
  fclose(in);
  }
else
  {
  if(!force_pd_rom)
    fprintf(stderr,"%s: no ROM found, using builtin PD `ROM'.\n",cmd_name);
  memcpy(x,pd_rom,sizeof(pd_rom));
  using_pd_rom=1;
  return;
  }

/* if enabled, change ROM colour-key messages to be less impenetrable.
 * XXX this won't work on the '150/'200
 */
if(!nc200 && !nc150 && alt_colour_msg)
  {
  static char *orig[6]=
    {
    "  YELLOW & RED   ",
    "      RED        ",
    " YELLOW & GREEN  ",
    "      GREEN      ",
    "  YELLOW & BLUE  ",
    "      BLUE       "
    };
  static char *changed[6]=
    {
    "  INSERT & LEFT  ",
    "      LEFT       ",
    " INSERT & RIGHT  ",
    "      RIGHT      ",
    "  INSERT & DOWN  ",
    "      DOWN       "
    };
  int f,addr,found;
  
  for(f=0;f<6;f++)
    {
    /* really sophisticated search algorithm :-) */
    for(addr=0x4000,found=0;addr<0x4400;addr++)
      if(mem[addr]==orig[f][0])
        if(strncmp((char *)mem+addr,orig[f],18)==0)
          strcpy((char *)mem+addr,changed[f]),found=1;
    
    if(!found)
      fprintf(stderr,"%s: couldn't change \"%s\" colour key message\n",
      	cmd_name,orig[f]);
    }
  }
}


void loadram(unsigned char *x)
{
#if 0	/* XXX not loaded */
FILE *in;
char *ram_file=libdir("nc100.ram");

if(nc150) ram_file=libdir("nc150.ram");
if(nc200) ram_file=libdir("nc200.ram");

if((in=fopen(ram_file,"rb"))!=NULL)
  {
  fread(x,1024,(nc150 || nc200)?128:64,in);
  fclose(in);
  }
else
  {
  fprintf(stderr,"%s: couldn't load RAM file `%s', using blank.\n",
  	cmd_name,ram_file);
  /* not fatal - blank it and carry on */
  memset(x,0,((nc150 || nc200)?128:64)*1024);
  }
#endif

memset(x,0,((nc150 || nc200)?128:64)*1024);
}


unsigned char *loadcard(void)
{
FILE *in;
unsigned char *cardmem;
char *card_file=libdir("nc100.card");

/* XXX single card used for all models. This is more complicated to handle,
 * but keeps required preload data size down.
 */
#if 0
if(nc150) card_file=libdir("nc150.card");
if(nc200) card_file=libdir("nc200.card");
#endif

/* we open read/write so mmap will work properly */
/* XXX except not with emscripten */
if((in=fopen(card_file,"r"))!=NULL)
  {
  fseek(in,0,SEEK_END);
  card_size=ftell(in)/1024;
#ifndef USE_MMAP_FOR_CARD
  rewind(in);
  if((cardmem=malloc(card_size*1024))==NULL)
    fprintf(stderr,"%s: out of memory.\n",cmd_name),exit(1);
  fread(cardmem,1024,card_size,in);
  fclose(in);
#else
  if((cardmem=mmap(0,card_size*1024,PROT_READ|PROT_WRITE,
  	MAP_SHARED,fileno(in),0))==MAP_FAILED)
    {
    fprintf(stderr,"%s: couldn't mmap() mem card file `%s'.\n",
    	cmd_name,card_file);
    exit(1);
    }
#endif /* USE_MMAP_FOR_CARD */
  }
else /* if fopen failed */
  {
  fprintf(stderr,"%s: couldn't open mem card file `%s'.\n",
  	cmd_name,card_file);
  cardmem=NULL;
  do_munmap=0;
  }

return(cardmem);
}


void writeram(unsigned char *x)
{
FILE *out;
char *ram_file=libdir("nc100.ram");

if(nc150) ram_file=libdir("nc150.ram");
if(nc200) ram_file=libdir("nc200.ram");

#if 0	/* XXX no RAM writing */
if((out=fopen(ram_file,"wb"))==NULL)
  fprintf(stderr,"%s: couldn't write RAM to `%s'.\n",cmd_name,ram_file);
else
  {
  fwrite(x,1024,(nc150 || nc200)?128:64,out);
  fclose(out);
  }
#endif
}


void writecard(unsigned char *x)
{
char *card_file=libdir("nc100.card");

if(nc150) card_file=libdir("nc150.card");
if(nc200) card_file=libdir("nc200.card");

/* we also close any open printer file while we're at it :-) */
if(printout) fclose(printout);

#if 0	/* XXX no card writing */
#ifdef USE_MMAP_FOR_CARD

if(do_munmap && munmap(x,card_size*1024)==-1)
  fprintf(stderr,"%s: couldn't munmap() mem card file `%s'.\n",
  	cmd_name,card_file);

#else
FILE *out;

if(card_size>0)
  {
  if((out=fopen(card_file,"wb"))==NULL)
    fprintf(stderr,"%s: couldn't write mem card to `%s'.\n",
    	cmd_name,card_file);
  else
    {
    fwrite(x,1024,card_size,out);
    fclose(out);
    }
  }
#endif /* !USE_MMAP_FOR_CARD */
#endif
}


/* I/O, quoting from nciospec.doc:
[we ignore some]
D0-DF                   RTC (TC8521)            R/W	(ours is read-only)
C0-C1                   UART (uPD71051)         R/W
B0-B9                   Key data in             R
A0                      Card Status etc.        R
90                      IRQ request status      R/W
70                      Power on/off control    W
60                      IRQ Mask                W
50-53                   Speaker frequency       W	(ignored)
40                      Parallel port data      W
30                      Baud rate etc.          W	(ignored)
20                      Card wait control       W	(ignored)
10-13                   Memory management       R/W
00                      Display memory start    W
*/


void do_paging(int page,int n)
{
lastpageout[page]=n;

switch(n&0xc0)
  {
  /* a good solid range of powers-of-two to be had here :-) */
  
  case 0x00:	/* ROM */
    if(nc150 || nc200)
      n&=0x1f;	/* 512k */
    else
      n&=0x0f;	/* 256k */
    memptr[page]=mem+16384*n;
    memattr[page]=0;
    break;
  case 0x40:	/* RAM */
    if(nc150 || nc200)
      n&=7;	/* 128k */
    else
      n&=3;	/* 64k */
    memptr[page]=mem+RAM_START+16384*n;
    memattr[page]=1;
    break;
  case 0x80:	/* memory card */
    n&=0x3f;	/* 1024k */
    memattr[page]=1;
    
    /* one way of detecting the size of a card requires
     * that writes which are off-card must fail, so...
     */
    if(16*n>=card_size) memattr[page]=0;
    
    /* for sub-1024k, limit addressed area to 512k/256k/0 */
    if(card_size<1024) n&=0x1f;	/* 512k */
    if(card_size<512)  n&=0x0f;	/* 256k */
    
    if(card_size<256)		/* 0 */
      memptr[page]=nocardmem;
    else
      memptr[page]=cardmem+16384*n;
    break;
  }
}


void serial_init(void)
{
struct termios new_tty_st;

if(!allow_serial) return;

tcgetattr(0,&old_tty_st);
tcgetattr(0,&new_tty_st);

/* equivalent of cfmakeraw(), or the old "stty raw -echo" */
new_tty_st.c_iflag&=~(IGNBRK | BRKINT | PARMRK | ISTRIP
                      | INLCR | IGNCR | ICRNL | IXON);
new_tty_st.c_oflag&=~OPOST;
new_tty_st.c_lflag&=~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
new_tty_st.c_cflag&=~(CSIZE | PARENB);
new_tty_st.c_cflag|=CS8;

tcsetattr(0,TCSAFLUSH,&new_tty_st);

/* make stdin non-blocking */
fcntl(0,F_SETFL,O_NONBLOCK);
}


void serial_uninit(void)
{
if(!allow_serial) return;

/* make stdin blocking again */
fcntl(0,F_SETFL,0);

/* restore old settings */
tcsetattr(0,TCSAFLUSH,&old_tty_st);
}


int serial_input_pending(void)
{
struct timeval tv;
fd_set fds;

if(!allow_serial) return(0);

tv.tv_sec=0; tv.tv_usec=0;
FD_ZERO(&fds); FD_SET(tty_fd,&fds);
if(select(tty_fd+1,&fds,NULL,NULL,&tv)<=0)
  return(0);

return(FD_ISSET(tty_fd,&fds));
}


int serial_output_allowed(void)
{
/* assume it always is. select() takes yonks in our terms. */
return(1);
}


int get_serial_byte(void)
{
unsigned char c=0;

if(!allow_serial) return(0);

if(read(tty_fd,&c,1)<=0)
  return(0);

return((int)c);
}


void put_serial_byte(int n)
{
unsigned char c=n;

#if 0
if(serial_output_allowed())
#endif
  {
  seroutbuf[seroutpos]=c;
  /* always leave a writable buffer byte, even if it drops chars */
  if(seroutpos+1<sizeof(seroutbuf))
    seroutpos++;
  }
}


void serout_flush(void)
{
if(!allow_serial) return;

if(seroutpos>0)
  write(tty_fd,seroutbuf,seroutpos);
seroutpos=0;
}


void put_printer_byte(int n)
{
static int printout_open_failed=0;
int bit=(nc200?1:4);

if(!printout && !printout_open_failed)
  {
  if((printout=fopen(libdir(PRINTER_OUTPUT),"ab"))==NULL)
    {
    printout_open_failed=1;	/* don't slow to crawl by trying every time */
    return;
    }
  }

fputc(n,printout);

/* now, the ROM (somewhat stupidly IMHO) requires the printer-ACK
 * interrupt in order to work. We emulate it in a similar way to
 * a serial-input int, except that in this case irq_status
 * has the relevant bit cleared here.
 */
if((irq_mask&bit)) irq_status&=~bit;
}


/* NC100 RTC code.
 *
 * The NC100 appears to use a TC8521AM.
 *
 * This emulation supports all four pages (time, alarm, and two spare
 * RAM pages of 13 4-bit values each). It has many limitations though;
 * it doesn't support:
 *
 * - alarms (any attempt to set one is ignored)
 * - setting the time
 * - 12hr operation (24hr only)
 * - the 1Hz and 16Hz timers
 * - the undocumented `test' register (reads as zero, writes ignored)
 */
unsigned int rtc_inout(int is_out,int h,int l,int a)
{
/* page3 is initialised with what early ROMs expect to see for time-set.
 * The pages are really 13 bytes, but 16 helps some code below look less
 * worrying. :-)
 */
/* XXX should probably be unsigned char like on the NC200 */
static int page2[16],page3[16]={6,7,1,1, 0,0,0,0, 0,0,0,0, 0,0,0,0};
static int page=0,page_reg=0;	/* page num and page register */
static struct tm *rtc_tm=NULL;
int pad=0xf0;	/* high nibble value */
time_t timet;

if(is_out)
  {
  switch(l)
    {
    case 0xdd:		/* page register */
      /* timer/alarm enable are ignored */
      page=(a&3);
      page_reg=a;
      
      /* we take the opportunity to read the time, assuming this port
       * is written to just before the time is read (to make sure the
       * time page is selected); anyway, both the ROM OS and ZCN do this,
       * so in practice it's good enough. :-)
       */
      timet=time(NULL);
      rtc_tm=localtime(&timet);
      break;
    
    case 0xde:		/* test register - ignored */
      break;
    
    case 0xdf:		/* reset register - ignored */
      break;
    
    default:		/* write to RAM */
      if(page<2)
        break;		/* ignore writes in time/alarm pages */
      if(page==2)
        page2[l&15]=a;
      else
        page3[l&15]=a;
    }
  
  return(0);
  }

/* otherwise it's an in. */

/* the chip deals in 4-bit values, and on an NC the top bits are
 * usually all set (but not always!), so we or with 0xf0.
 */
if(l==0xdd)
  return(pad|page_reg);

if(l==0xde || l==0xdf)	/* write-only, ret 0 */
  return(pad);

switch(page)
  {
  case 0:		/* time page */
    /* XXX from old TODO (is this still true, or was it an NC200 thing?):
     * Need to fix RTC. My kludge which grabbed rtc_tm isn't being triggered
     * any more, so the time doesn't work.
     */
    if(!rtc_tm) return(pad);
    
    switch(l)
      {
      case 0xd0:	return(pad|(rtc_tm->tm_sec%10));
      case 0xd1:	return(pad|(rtc_tm->tm_sec/10));
      case 0xd2:	return(pad|(rtc_tm->tm_min%10));
      case 0xd3:	return(pad|(rtc_tm->tm_min/10));
      case 0xd4:	return(pad|(rtc_tm->tm_hour%10));
      case 0xd5:	return(pad|(rtc_tm->tm_hour/10));
      case 0xd6:	return(pad|(rtc_tm->tm_wday));
      case 0xd7:	return(pad|(rtc_tm->tm_mday%10));
      case 0xd8:	return(pad|(rtc_tm->tm_mday/10));
      case 0xd9:	return(pad|((rtc_tm->tm_mon+1)%10));
      case 0xda:	return(pad|((rtc_tm->tm_mon+1)/10));
      /* XXX not clear what would be best to do for years past 2099 */
      case 0xdb:	return(pad|((rtc_tm->tm_year-90)%10));
      case 0xdc:	return(pad|((rtc_tm->tm_year-90)/10));
      default:		return(pad);
      }
  
  case 1:		/* alarm page */
    switch(l)
      {
      case 0xda:
        return(pad|1);	/* set low bit since we always use 24hr */
      case 0xdb:
        if(!rtc_tm) return(pad);		/* XXX see earlier note */
        return(pad|(rtc_tm->tm_year&3));	/* good until 2099 */
      case 0xdc:
        /* this identifies to a clueful program that it's running
         * under nc100em, as needed for the co-operative mouse support
         * `emulation'. This is an otherwise-unused (and useless)
         * byte, so this shouldn't break anything.
         *
         * (...20+ years pass...)
         *
         * And it didn't! Except on the NC200, where doing a check for
         * this (on a real one, which decodes port DDh as D1h) just
         * happens to mean that you end up writing a random RTC memory
         * location. FFS. So this is still here, and still works on
         * NC100/NC150, but is no longer documented and must not be
         * checked for on an NC200.
         *
         * (If you must do the check on an NC200, the least-bad way is to
         * do "xor a/out (0deh),a" before the "ld a,9/out (0ddh),a", which
         * should harmlessly zero the RTC test register on NC100/NC150, and
         * on a real NC200 selects the clock seconds, so you'd only be
         * messing with that. So as I say... least-bad.)
         */
        return(100);	/* as in NC~ :-) */
      default:
        return(pad);	/* otherwise zero */
      }

  case 2:
    return(pad|page2[l&15]);
  
  case 3:
    return(pad|page3[l&15]);
  }

/* can't get here, but just in case... */
return(pad|0xf);
}


/* NC200 RTC code.
 *
 * The NC200 seems to use the functional equivalent of an MC146818
 * or similar, at time of writing see e.g.:
 *
 *   https://www.nxp.com/docs/en/data-sheet/MC146818.pdf
 *
 * This emulation supports a read-only time and date in binary only,
 * and emulates the extra RAM (there are dozens of bytes! dozens!).
 * It doesn't support:
 *
 * - BCD mode
 * - alarms (any attempt to set one is ignored)
 * - setting the time
 * - 12hr operation (24hr only)
 * - interrupts
 * - square-wave generation
 * - any other random junk
 * 
 * It's enough for the ROM's time/date-reading to work, and should
 * be sufficient for NC200 RTC support in ZCN.
 *
 * One mystifying issue - how does this RTC cope with leap years? The
 * MC146818 documentation suggests two-digit years implicitly in the
 * range of 1900-1999, but the ROM software seems to support years
 * from 1920 to 2099 (which keeps leap years at strictly every 4
 * years). Meanwhile, the ROM treats year zero on the actual ports as
 * 1990. (!)
 *
 * Some notes from testing a real NC200:
 *
 * - At least on the NC200 I've used, the answer to the leap-year
 *   question above is "it doesn't". If you let the RTC rollover from
 *   28 Feb after setting time/date with the ROM software, it acts
 *   like 1994 was a leap year, and 1996 wasn't, and every two years
 *   it will add or lose a day in February when it rolls over (it can
 *   be manually set to 29 Feb or 1 Mar as required after that
 *   happens, but that's all you can do to fix it). Luckily, this can
 *   be ignored in nc100em - the real time/date can just be made
 *   available.
 *
 * - The "extra" RAM (location 14 onwards) seemed to be all FFh, and
 *   couldn't actually be changed. I don't know if this is universal,
 *   or just a RAM failure on a nearly-30-year-old chip. Though I do
 *   note that the bq3285 (a clone from Texas Instruments) is
 *   documented as having a RAM-reset pin with this precise effect,
 *   explicitly setting all the extra RAM to FFh. Probably
 *   coincidental though.
 */
unsigned int rtc200_inout(int is_out,int h,int l,int a)
{
static time_t inittime=0;
static time_t timeofs=0;	  /* time offset, for testing */
static unsigned char rtcram[64];  /* only bytes 14..63 fully available */
static struct tm *rtc_tm=NULL;
static int reg=0;
time_t timet;

if(!inittime) inittime=time(NULL);

if(l<0xd0 || l>0xd1) return(0);

/* select register/RAM address */
if(is_out && l==0xd0)
  {
  reg=(a&0x3f);
  return(0);
  }

if(l==0xd0) return(0);

/* so, l==0xd1 - read/write current register/RAM */
if(is_out)
  {
  /* write reg A/B/C/D and spare RAM only */
  /* XXX could maybe allow "gaps" too, like ROM abuse of day-of-week */
  if(reg>9)
    rtcram[reg]=a;
  return(0);
  }

/* otherwise it's an in. */

switch(reg)
  {
  /* 1,3,5 are alarm-related, they'll just return zero */
  case 0:	timet=time(NULL);
  		if(timeofs) timet=timet-inittime+timeofs;
  		rtc_tm=localtime(&timet);
  		return(rtc_tm?rtc_tm->tm_sec:0);
  case 2:	return(rtc_tm?rtc_tm->tm_min:0);
  case 4:	return(rtc_tm?rtc_tm->tm_hour:0);
  case 7:	return(rtc_tm?rtc_tm->tm_mday:0);
  case 8:	return(rtc_tm?(rtc_tm->tm_mon+1):0);
  /* XXX would 2090-2099 work on this? And what to do for >2099? */
  case 9:	return(rtc_tm?(rtc_tm->tm_year-90):0);

  case 0xa:
    /* reg A, UIP always zero, i.e. ok to read time; in theory this
     * ought be a good place to set rtc_tm, but it doesn't work out in
     * practice.
     *
     * (I think that's because the ROM will often read seconds only,
     * without checking UIP at all, using a change in seconds to see
     * whether it needs to reread the time/date. This may be
     * "cheating" on the ROM's part :-) but there's some logic to it
     * if you ignore what the (original chip's) datasheet says to do
     * and just assume that a seconds update is atomic and ok to read
     * at any time.)
     */
    return((rtcram[0xa]&127));
    
  case 0xb:
    /* reg B is just pegged to a reasonable value */
    return(6);		/* binary mode, 24hr */

  case 0xc:
    /* reg C */
    return(0);		/* interrupt-related, not supported */
    
  case 0xd:
    /* reg D */
    return(128);	/* VRT bit set */
    
    /* The VRT description I read was quite confusingly phrased, but
     * in practice I know the NC200 wants it to be set.
     */
  
  default:
    return(rtcram[reg]);
  }

/* can't get here, but just in case... */
return(0);
}



unsigned int in(int h,int l)
{
static int ts=0;	/* extra t-states for this out, times 256 */
static int mx=0,my=0,mb=0;

/* h is ignored by the NC100 */
switch(l)
  {
  case 0xe0:
#ifdef NC200_DISK_SUPPORT
    if(nc200)
      return(ts|fdc_read_ctrl(nc200_fdc));
    else
#endif
      return(ts|255);
    
  case 0xe1:
#ifdef NC200_DISK_SUPPORT
    if(nc200)
      return(ts|fdc_read_data(nc200_fdc));
    else
#endif
      return(ts|255);
    
  /* RTC */
  case 0xd0: case 0xd1: case 0xd2: case 0xd3:
  case 0xd4: case 0xd5: case 0xd6: case 0xd7:
  case 0xd8: case 0xd9: case 0xda: case 0xdb:
  case 0xdc: case 0xdd: case 0xde: case 0xdf:
    return(ts|(nc200?rtc200_inout(0,h,l,0):rtc_inout(0,h,l,0)));
  
  /* UART */
  case 0xc0:	/* this reads from the serial port */
    return(ts|get_serial_byte());
  
  case 0xc1:    /* returns bit 0=1 if can send a byte */
    /* we assume it's always possible */
    return(ts|1);
    
  /* keyboard */
  case 0xb0: case 0xb1: case 0xb2: case 0xb3: case 0xb4:
  case 0xb5: case 0xb6: case 0xb7: case 0xb8: case 0xb9:
    /* reading 0xb9 also sets bit 3 of irq_status on NC100 */
    if(!nc200 && l==0xb9)
      irq_status|=8;
    return(ts|keyports[l-0xb0]);
  
  case 0xa0:	/* card etc. status */
    return(ts|card_status);
  
  case 0x90:	/* IRQ status */
    return(ts|irq_status);
  
  case 0x80:	/* NC200 printer busy */
    if(nc200) return(ts);	/* zero */
    return(ts|255);
  
  /* memory paging */
  case 0x10: return(ts|lastpageout[0]);
  case 0x11: return(ts|lastpageout[1]);
  case 0x12: return(ts|lastpageout[2]);
  case 0x13: return(ts|lastpageout[3]);
  
  /* co-operative mouse `emulation' - doesn't emulate a mouse, but does
   * provide a way to read the pointer location for nc100em-aware programs.
   * zcnpaint (and zcnlib) as of ZCN 1.3 support this, for example.
   */
  case 0x2a:
    {
    /* grab current state of pointer, and read buttons. You should read
     * this port before reading the other three.
     */
    static struct timeval last_read={0,0};
    struct timeval now;
    int old,new;
    int l,m,r;
    
    /* limit it to max of 500 times/sec - without this you could really
     * give your X server hell. :-)
     */
    gettimeofday(&now,NULL);
    old=last_read.tv_usec;
    new=now.tv_usec;
    if(now.tv_sec>last_read.tv_sec)
      new+=1000000;
    if(new-old<=2000)
      return(ts|mb);
    
    last_read.tv_sec=now.tv_sec;
    last_read.tv_usec=now.tv_usec;
    
    get_mouse_state(&mx,&my,&mb);
    
    l=( mb    &1);
    m=((mb>>1)&1);
    r=((mb>>2)&1);
    
    /* there are multiple bit layouts in the returned byte, to allow
     * the most usual bit layouts to be read directly without needing
     * the caller to do hairy bit-twiddling. The byte layout is:
     *
     * 76543210
     * 0lmrmlrl  ...with l=left, m=middle, r=right mouse buttons.
     */
    mb=((l<<6)|(m<<5)|(r<<4)|(m<<3)|(l<<2)|(r<<1)|l);
    
    return(ts|mb);
    }
  
  case 0x2b:	/* read x low */
    return(ts|(mx&255));
  
  case 0x2c:	/* read x high */
    return(ts|(mx>>8));
  
  case 0x2d:	/* read y low */
    return(ts|my);

  case 0x2f:	/* check NC model - used by PD ROM */
    if(nc200)
      return(ts|200);
    else
      {
      if(nc150)
        return(ts|150);
      else
        return(ts|100);
      }
  }

/* otherwise... */
return(ts|255);
}


unsigned int out(int h,int l,int a)
{
static int printer_byte=0;	/* data output to port 0x40 */
static int printer_strobe=1;	/* normal state is high */
static int ts=0;		/* extra t-states for this out */

/* h is ignored by the NC100 */
switch(l)
  {
#ifdef NC200_DISK_SUPPORT
  case 0xe1:
    if(nc200)
      fdc_write_data(nc200_fdc,a);
    return(ts);
#endif
  
  /* RTC */
  case 0xd0: case 0xd1: case 0xd2: case 0xd3:
  case 0xd4: case 0xd5: case 0xd6: case 0xd7:
  case 0xd8: case 0xd9: case 0xda: case 0xdb:
  case 0xdc: case 0xdd: case 0xde: case 0xdf:
    nc200?rtc200_inout(1,h,l,a):rtc_inout(1,h,l,a);
    return(ts);
  
  /* UART */
  case 0xc0:	/* this writes to the serial port */
    put_serial_byte(a);
    return(ts);
  
  case 0xc1:    /* sets up various serial parms which we ignore */
    return(ts);
  
  case 0x90:	/* IRQ status */
    /* when a zero is written to a bit, it should re-enable that IRQ line */
    irq_status|=(a^255);
    return(ts);
  
  case 0x70:	/* power on/off */
    /* if bit 0 is 0, turn off */
    if((a&1)==0)
      {
      write_ram_on_close=1;	/* be sure to save RAM, for context-saving */
      exit_program();
      }
    
    /* on NC200, bit 2 is backlight (0=on). */
    
    return(ts);
  
  case 0x60:	/* set irq mask */
    irq_mask=a;
    return(ts);
  
  case 0x50: case 0x51: case 0x52: case 0x53:	/* sound */
    return(ts);
  
  case 0x40:	/* parallel port data */
    printer_byte=a;	/* not written until strobe is pulsed */
    return(ts);
  
  case 0x30:	/* baud rate etc., ignored except for printer strobe */
    /* bit 6 is the strobe - low then high means print byte. */
    if(!printer_strobe && (a&0x40))
      put_printer_byte(printer_byte);
    printer_strobe=(a&0x40);
#ifdef NC200_DISK_SUPPORT
    if(nc200)
      fdc_set_motor(nc200_fdc,!(a&32));
#endif
    return(ts);
  
  case 0x20:	/* card wait control, ignored */
#ifdef NC200_DISK_SUPPORT
    if(nc200)
      fdc_set_terminal_count(nc200_fdc,a&1);
#endif
    return(ts);
  
  /* memory paging */
  case 0x10: do_paging(0,a); return(ts);
  case 0x11: do_paging(1,a); return(ts);
  case 0x12: do_paging(2,a); return(ts);
  case 0x13: do_paging(3,a); return(ts);
  
  case 0:	/* set screen addr */
    /* on NC200 only high 3 bits are significant */
    scrnaddr=(a&(nc200?0xe0:0xf0))*256;
    force_full_redraw=1;	/* override differential updating */
    return(ts);
  }

/* otherwise... */
return(ts);
}



void sighandler(int a)
{
signal_int_flag=1;
}


Uint32 sighandler_sdl(Uint32 interval,void *param)
{
/* XXX doesn't attempt to fix timer crapitude */
signal_int_flag=1;

return(interval);
}


/* only actually called if NC200 disk support used,
 * but always defined.
 */
void nc200_fdc_tick(void)
{
#ifdef NC200_DISK_SUPPORT
fdc_tick(nc200_fdc);
#endif
}


#ifdef NC200_DISK_SUPPORT

void nc200_disk_int_set(FDC_PTR self,int status)
{
diskint=status;
}


void err765(int debuglevel,char *fmt,va_list ap)
{
/* just ignore them all, for now */
}


void disk_init(void)
{
/* XXX hacky - really this shouldn't assume that the disk image will
 * still be there the next time it's opened just because it was the
 * first time. Doing it via err765 would be a pain, though.
 */
FILE *in=fopen(libdir(NC200_DISK_IMAGE),"rb");
int makenew=1;

if(in)
  fclose(in),makenew=0;

lib765_register_error_function(err765);

nc200_fdc=fdc_new();
nc200_drive=fd_newdsk();
nc200_drive_dummy=fd_new();
fd_settype(nc200_drive,FD_35);
fd_setheads(nc200_drive,2);
fd_setcyls(nc200_drive,80);
fdd_setfilename(nc200_drive,libdir(NC200_DISK_IMAGE));

if(makenew)
  {
  fprintf(stderr,"%s: couldn't open NC200 disk image `%s', created blank.\n",
  	cmd_name,NC200_DISK_IMAGE);
  fprintf(stderr,"\t(Be sure to format this image in the emulated NC200 before\n");
  fprintf(stderr,"\ttrying to use it.)\n");
  fdd_new_dsk(nc200_drive);
  }

fdc_reset(nc200_fdc);
fdc_setisr(nc200_fdc,nc200_disk_int_set);
fdc_setdrive(nc200_fdc,0,nc200_drive);
fdc_setdrive(nc200_fdc,1,nc200_drive_dummy);
fdc_setdrive(nc200_fdc,2,nc200_drive_dummy);
fdc_setdrive(nc200_fdc,3,nc200_drive_dummy);
fdc_write_drr(nc200_fdc,1);	/* DD */

fd_setreadonly(nc200_drive,0);
}


void disk_uninit(void)
{
fdc_destroy(&nc200_fdc);
fd_destroy(&nc200_drive);
fd_destroy(&nc200_drive_dummy);
}

#endif /* NC200_DISK_SUPPORT */


/* init memory etc. */
void common_init(void)
{
loadrom(mem);
loadram(mem+RAM_START);
cardmem=loadcard();
set_card_status();

/* XXX need to check for '200 compat (eh?) */

/* load file at 100h if they specified one. */
if(binfile_name)
  {
  FILE *in;
  
  /* only if it's the PD ROM */
  if(!using_pd_rom)
    fprintf(stderr,
    	"%s: can only load a file to boot with PD ROM (try `-p').\n",
    	cmd_name);
  else
    {
    /* blast the context-save magic so it jumps to 100h */
    memset(mem+RAM_START+0xb200,0,4);
    
    if((in=fopen(binfile_name,"rb"))==NULL)
      fprintf(stderr,"%s: couldn't open file `%s'.\n",cmd_name,binfile_name);
    else
      fread(mem+RAM_START+0x100,1,49152,in),fclose(in);
    }
  }

serial_init();

#ifdef NC200_DISK_SUPPORT
if(nc200)
  disk_init();
#endif
}


void common_uninit(void)
{
serial_uninit();
#ifdef NC200_DISK_SUPPORT
if(nc200)
  disk_uninit();
#endif
}


static void exit_program_flag_set(int foo)
{
exit_program_flag=1;
}


/* must be called after common_init(), not before */
void startsigsandtimer(void)
{
if(wait_for_sec)
  sleep(2);
}


void usage_help(void)
{
printf(
  "%s " NC100EM_VER " - copyright (C) 1995-2023 Russell Marks.\n"
  "Z80 emulation and Xlib code from xz80, copyright (C) 1994 Ian Collier.\n"
  "\n",
  cmd_name);
printf("%s is the web-based version of nc100em.\n\n",cmd_name);

printf("usage: %s [-125hmps] [-r refresh_rate] [-S scale] "
	"[file_to_boot.bin]\n",cmd_name);
puts("\n"
"	-1	emulate an NC100 (the usual default).\n"
"	-2	emulate an NC200.\n"
"	-5	emulate an NC150.\n"
"	-h	this usage help.\n"
"	-i	invert graphics (swap black and white).\n"
"	-m	modify `colour' key messages in ROM, to make them more\n"
"		understandable when using nc100em. For example, with `-m',\n"
"		nc100em changes \"YELLOW & RED\" to \"INSERT & LEFT\".\n"
"		(Only supported in NC100 mode on UK ROMs.)\n"
"	-p	use the builtin PD `ROM' even if a ROM file is present.\n"
"		Useful if you want to boot a file.\n"
"	-r	set how often the screen is redrawn in 1/100ths of a second\n"
"		(1 by default).\n"
"	-s	enable serial I/O on stdin/stdout. Only relevant to\n"
"		graphical versions (always enabled in tnc100em).\n"
"	-S	set scaling factor in range 1..16 (only has an effect in\n"
"		dnc100em/gnc100em), 3 by default. Large factors may be slow.\n"
"	-w	wait for two seconds before starting the Z80, to give you\n"
"		time to hold down a few keys (for resets etc.).\n"
"\n"
"  file_to_boot.bin	load file into RAM at 100h and run it. The file must\n"
"			start with F3h (`di'). (This is generally used to\n"
"			boot ZCN (zcn.bin) when using the builtin `ROM'.)\n"
"\n"
"If the emulator's command name ends in `200em' or `150em', it will default\n"
"to running in NC200 or NC150 mode respectively.\n");
}


void parseoptions(int argc,char *argv[])
{
int done=0;

/* if command name ends in "200em" or "150em", default to
 * emulating NC200 or NC150 respectively.
 */
if(argc>=1 && argv[0] && strlen(argv[0])>=5)
  {
  if(strcmp(argv[0]+strlen(argv[0])-5,"200em")==0)
    nc200=1;
  else
    if(strcmp(argv[0]+strlen(argv[0])-5,"150em")==0)
      nc150=1;
  }

opterr=0;

do
  switch(getopt(argc,argv,"125himpr:sS:w"))
    {
    case '1':
      nc200=0; nc150=0;
      break;
    case '2':
      nc200=1; nc150=0;
      break;
    case '5':
      nc200=0; nc150=1;
      break;
    case 'h':
      usage_help();
      exit(1);
    case 'i':
      invert_video=1;
      break;
    case 'm':
      alt_colour_msg=1;
      break;
    case 'p':	/* force PD ROM even if nc100.rom is present */
      force_pd_rom=1;
      break;
    case 'r':	/* refresh rate */
      scrn_freq=atoi(optarg);
      if(scrn_freq<1) scrn_freq=1;
      if(scrn_freq>100) scrn_freq=100;
      break;
    case 's':	/* enable serial emulation (only affects [gx]nc100em) */
      allow_serial=1;
      break;
    case 'S':	/* set scale factor */
      default_scale=atoi(optarg);
      if(default_scale<1) default_scale=1;
      if(default_scale>16) default_scale=16;
      break;
    case 'w':	/* wait */
      wait_for_sec=1;
      break;
    case '?':
      switch(optopt)
        {
        case 'r':
          fprintf(stderr,"%s: "
          	"the -r option needs a refresh rate argument.\n",
                cmd_name);
          break;
        case 'S':
          fprintf(stderr,"%s: "
          	"the -S option needs a scaling factor argument.\n",
                cmd_name);
          break;
        default:
          fprintf(stderr,"%s: "
          	"option `%c' not recognised.\n",
                cmd_name,optopt);
        }
      exit(1);
    case -1:
      done=1;
    }
while(!done);

if(optind==argc-1)	/* if a filename given... */
  binfile_name=argv[optind];

/* for emscripten web version, only really sensible to boot ZCN.
 * boot.bin boots from the card, others just use the boot-file method
 * instead.
 */
binfile_name=(nc200?"zcn200.bin":(nc150?"zcn150.bin":"boot.bin"));
force_pd_rom=1;

/* change scrnaddr if NC200 */
if(nc200) scrnaddr=0xe000;
}
